import platform
import json
import os

def parseSingeRule(rule,options):
    #Parse a rule from the versions.json
    if rule["action"] == "allow":
        returnvalue = False
    elif rule["action"] == "disallow":
        returnvalue = True
    if "os" in rule:
        for key, value in rule["os"].items():
            if key == "name":
                if value == "windows" and platform.system() != 'Windows':
                    return returnvalue
                elif value == "osx" and platform.system() != 'Darwin':
                    return returnvalue
                elif value == "linux" and platform.system() != 'Linux':
                    return returnvalue
            elif key == "arch":
                if value == "x86" and platform.architecture()[0] != "32bit":
                    return returnvalue
    if "features" in rule:
        for key, value in rule["features"].items():
            if key == "has_custom_resolution" and not options.get("customResolution",False):
                return returnvalue
            elif key == "is_demo_user" and not options.get("demo",False):
                return returnvalue
    return not returnvalue

def parseRuleList(data,ruleString,options):
    #Parse a rule list
    if not ruleString in data:
        return True
    for i in data[ruleString]:
        if not parseSingeRule(i,options):
            return False
    return True

def getNatives(data):
    if platform.architecture()[0] == "32bit":
        arch_type = "32"
    else:
        arch_type = "64"
    if "natives" in data:
        if platform.system() == 'Windows':
            if "windows" in data["natives"]:
                return data["natives"]["windows"].replace("${arch}",arch_type)
            else:
                return ""
        elif platform.system() == 'Darwin':
            if "osx" in data["natives"]:
                return data["natives"]["osx"].replace("${arch}",arch_type)
            else:
                return ""           
        else:
            if "linux" in data["natives"]:
                return data["natives"]["linux"].replace("${arch}",arch_type)
            else:
                return "" 
    else:
        return ""

def inherit_json(original_data,path):
    #See https://github.com/tomsik68/mclauncher-api/wiki/Version-Inheritance-&-Forge
    inherit_version = original_data["inheritsFrom"]
    with open(os.path.join(path,"versions",inherit_version,inherit_version + ".json")) as f:
        new_data = json.load(f)
    for key, value in original_data.items():
        if isinstance(value,list) and isinstance(new_data.get(key,None),list):
            new_data[key] = value + new_data[key]
        elif isinstance(value,dict) and isinstance(new_data.get(key,None),dict):
            for a, b in value.items():
                if isinstance(b,list):
                    new_data[key][a] = new_data[key][a] + b
        else:
            new_data[key] = value
    return new_data



from .natives import extract_natives_file, get_natives
from .helper import parseRuleList, inherit_json
from .utils import get_library_version
import requests
import shutil
import json
import os

def empty(arg):
    pass

def download_file(url,path,callback):
    if os.path.isfile(path):
        return
    try:
        os.makedirs(os.path.dirname(path))
    except:
        pass
    callback.get("setStatus",empty)("Download " + os.path.basename(path))
    r = requests.get(url, stream=True, headers={"user-agent": "minecraft-launcher-lib/" + get_library_version()})
    with open(path, 'wb') as f:
        r.raw.decode_content = True
        shutil.copyfileobj(r.raw, f)

def install_libraries(data,path,callback):
    callback.get("setMax",empty)(len(data["libraries"]))
    for count, i in enumerate(data["libraries"]):
        #Check, if the rules allow this lib for the current system
        if not parseRuleList(i,"rules",{}):
            continue
        #Turn the name into a path
        currentPath = os.path.join(path,"libraries")
        libPath, name, version = i["name"].split(":")
        for l in libPath.split("."):
            currentPath = os.path.join(currentPath,l)
        currentPath = os.path.join(currentPath,name,version)
        native = get_natives(i)
        #Check if there is a native file
        if native != "":
            jarFilenameNative = name + "-" + version + "-" + native + ".jar"
        jarFilename = name + "-" + version + ".jar"
        if not "downloads" in i:
            if "extract" in i:
                extract_natives_file(os.path.join(currentPath,jarFilenameNative),os.path.join(path,"versions",data["id"],"natives"),i["extract"])
            continue
        if "artifact" in i["downloads"]:
            download_file(i["downloads"]["artifact"]["url"],os.path.join(currentPath,jarFilename),callback)
        if native != "":
            download_file(i["downloads"]["classifiers"][native]["url"],os.path.join(currentPath,jarFilenameNative),callback)
            if "extract" in i:
                extract_natives_file(os.path.join(currentPath,jarFilenameNative),os.path.join(path,"versions",data["id"],"natives"),i["extract"])
        callback.get("setProgress",empty)(count)

def install_assets(data,path,callback):
    #Old versions dosen't have this
    if not "assetIndex" in data:
        return
    #Download all assets
    download_file(data["assetIndex"]["url"],os.path.join(path,"assets","indexes",data["assets"] + ".json"),callback)
    with open(os.path.join(path,"assets","indexes",data["assets"] + ".json")) as f:
        assets_data = json.load(f)
    #The assets gas a hash. e.g. c4dbabc820f04ba685694c63359429b22e3a62b5
    #With this hash, it can be download from https://resources.download.minecraft.net/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    #And saved at assets/objects/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    callback.get("setMax",empty)(len(assets_data["objects"]))
    count = 0
    for key,value in assets_data["objects"].items():
        download_file("https://resources.download.minecraft.net/" + value["hash"][:2] + "/" + value["hash"],os.path.join(path,"assets","objects",value["hash"][:2],value["hash"]),callback)
        count += 1
        callback.get("setProgress",empty)(count)

def do_version_install(versionid,path,callback,url=None):
    #Download and read versions.json
    if url:
        download_file(url,os.path.join(path,"versions",versionid,versionid + ".json"),callback)
    with open(os.path.join(path,"versions",versionid,versionid + ".json")) as f:
        versiondata = json.load(f)
    #For Forge
    if "inheritsFrom" in versiondata:
        versiondata = inherit_json(versiondata,path)
    install_libraries(versiondata,path,callback)
    install_assets(versiondata,path,callback)
    #Download minecraft.jar
    if "downloads" in versiondata:
        download_file(versiondata["downloads"]["client"]["url"],os.path.join(path,"versions",versiondata["id"],versiondata["id"] + ".jar"),callback)

def install_minecraft_version(versionid,path,callback=None):
    if callback == None:
        callback = {}
    version_list = requests.get("https://launchermeta.mojang.com/mc/game/version_manifest.json").json()
    for i in version_list["versions"]:
        if i["id"] == versionid:
            do_version_install(versionid,path,callback,url=i["url"])
            return True
    if not os.path.isdir(os.path.join(path,"versions")):
        return False
    for i in os.listdir(os.path.join(path,"versions")):
        if i == versionid:
            do_version_install(versionid,path,callback)
            return True
    return False



#Thanks to https://github.com/Polyr/mojang-api/blob/master/mojang_api/servers/authserver.py
import requests
import uuid

def login_user(username, password):
    payload = {
        "agent": {
            "name": "Minecraft",
            "version": 1
        },
        "username": username,
        "password": password,
        "clientToken": uuid.uuid4().hex
    }

    response = requests.post("https://authserver.mojang.com/authenticate", json=payload)
    return response.json()

def validate_access_token(access_token):
    payload = {
        "accessToken": access_token,
    }
    response = requests.post("https://authserver.mojang.com/validate", json=payload)
    return response.status_code == 204

def refresh_access_token(access_token, client_token):
    payload = {
        'accessToken': access_token,
        'clientToken': client_token
    }

    response = requests.post("https://authserver.mojang.com/refresh", json=payload)
    return response.json()

def logout_user(username, password):
    payload = {
        'username': username,
        'password': password
    }
    response = requests.post("https://authserver.mojang.com/signout", json=payload)
    return  response.status_code == 204

def invalidate_access_token(access_token, client_token):
    payload = {
        'accessToken': access_token,
        'clientToken': client_token
    }
    response = requests.post("https://authserver.mojang.com/invalidate", json=payload)
    return response

def upload_skin(uuid, access_token, path, slim=False):
    headers = {
        "Authorization": "Bearer " + access_token
    }
    files = {
        "model": "slim" if slim else "",
        "file": open(path,"rb")
    }
    response = requests.put("https://api.mojang.com/user/profile/{uuid}/skin".format(uuid=uuid), headers=headers, files=files)
    return response

def reset_skin(uuid, access_token):
    headers = {
        "Authorization": "Bearer " + access_token
    }
    response = requests.delete("https://api.mojang.com/user/profile/{uuid}/skin".format(uuid=uuid),headers=headers)
    return response



from .helper import parseRuleList, inherit_json
from .utils import get_library_version
from .natives import get_natives
import platform
import json
import copy
import os

def get_libraries(data,path):
    if platform.system() == "Windows":
        classpath_seperator = ";"
    else:
        classpath_seperator = ":"
    libstr = ""
    for i in data["libraries"]:
        if not parseRuleList(i,"rules",{}):
            continue
        currentPath = os.path.join(path,"libraries")
        libPath, name, version = i["name"].split(":")
        for l in libPath.split("."):
            currentPath = os.path.join(currentPath,l)
        currentPath = os.path.join(currentPath,name,version)
        native = get_natives(i)
        if native == "":
            jarFilename = name + "-" + version + ".jar"
        else:
            jarFilename = name + "-" + version + "-" + native + ".jar"
        currentPath = os.path.join(currentPath,jarFilename)
        libstr = libstr + currentPath + classpath_seperator
    if "jar" in data:
        libstr = libstr + os.path.join(path,"versions",data["jar"],data["jar"] + ".jar")
    else:
        libstr = libstr + os.path.join(path,"versions",data["id"],data["id"] + ".jar")
    return libstr

def replace_arguments(argstr,versionData,path,options):
    #Replace all arguments with the needed value
    argstr = argstr.replace("${natives_directory}",options["nativesDirectory"])
    argstr = argstr.replace("${launcher_name}",options.get("launcherName","minecraft-launcher-lib"))
    argstr = argstr.replace("${launcher_version}",options.get("launcherVersion",get_library_version()))
    argstr = argstr.replace("${classpath}",options["classpath"])
    argstr = argstr.replace("${auth_player_name}",options.get("username","{username}"))
    argstr = argstr.replace("${version_name}",versionData["id"])
    argstr = argstr.replace("${game_directory}",options.get("gameDirectory",path))
    argstr = argstr.replace("${assets_root}",os.path.join(path,"assets"))
    argstr = argstr.replace("${assets_index_name}",versionData.get("assets",versionData["id"]))
    argstr = argstr.replace("${auth_uuid}",options.get("uuid","{uuid}"))
    argstr = argstr.replace("${auth_access_token}",options.get("token","{token}"))
    argstr = argstr.replace("${user_type}","mojang")
    argstr = argstr.replace("${version_type}",versionData["type"])
    argstr = argstr.replace("${user_properties}","{}")
    argstr = argstr.replace("${resolution_width}",options.get("resolutionWidth","854"))
    argstr = argstr.replace("${resolution_height}",options.get("resolutionHeight","480"))
    argstr = argstr.replace("${game_assets}",os.path.join(path,"assets","virtual","legacy"))
    argstr = argstr.replace("${auth_session}",options.get("token","{token}"))
    return argstr

def get_arguments_string(versionData,path,options):
    arglist = []
    for v in versionData["minecraftArguments"].split(" "):
        v = replace_arguments(v,versionData,path,options)
        arglist.append(v)
    #Custom resolution is not in the list
    if options.get("customResolution",False):
        arglist.append("--width")
        arglist.append(options.get("resolutionWidth","854"))
        arglist.append("--height")
        arglist.append(options.get("resolutionHeight","480"))
    if options.get("demo",False):
        arglist.append("--demo")
    return arglist

def get_arguments(data,versionData,path,options):
    arglist = []
    for i in data:
        #Rules might has 2 different names in different versions.json
        if not parseRuleList(i,"compatibilityRules",options):
            continue
        if not parseRuleList(i,"rules",options):
            continue
        #i could be the argument
        if isinstance(i,str):
            arglist.append(replace_arguments(i,versionData,path,options))
        else:
            #Sometimes  i["value"] is the argument
            if isinstance(i["value"],str):
                arglist.append(replace_arguments(i["value"],versionData,path,options))
            #Sometimes i["value"] is a list of arguments
            else:
                for v in i["value"]:
                    v = replace_arguments(v,versionData,path,options)
                    arglist.append(v)
    return arglist

def get_minecraft_command(version,path,options):
    options = copy.copy(options)
    with open(os.path.join(path,"versions",version,version + ".json")) as f:
        data = json.load(f)
    if "inheritsFrom" in data:
        data = inherit_json(data,path)
    options["nativesDirectory"] = options.get("nativesDirectory",os.path.join(path,"versions",data["id"],"natives"))
    options["classpath"] = get_libraries(data,path)
    command = [options.get("executablePath","java")]
    if "jvmArguments" in options:
        command = command + options["jvmArguments"]
    #Newer Versions have jvmArguments in version.json
    if isinstance(data.get("arguments",None),dict):
        if "jvm" in data["arguments"]:
            command = command + get_arguments(data["arguments"]["jvm"],data,path,options)
        else:
            command.append("-Djava.library.path=" + options["nativesDirectory"])
            command.append("-cp")
            command.append(options["classpath"])
    else:
        command.append("-Djava.library.path=" + options["nativesDirectory"])
        command.append("-cp")
        command.append(options["classpath"])
    command.append(data["mainClass"])
    if "minecraftArguments" in data:
        #For older versions
        command = command + get_arguments_string(data,path,options)
    else:
        command = command + get_arguments(data["arguments"]["game"],data,path,options)
    if "server" in options:
        command.append("--server")
        command.append(options["server"])
        if "port" in options:
            command.append("--port")
            command.append(options["port"])
    return command
